/* RealSpeak.cpp

RealSpeak(TM) driver.

Part of the caiviar package. 

Copyright (c) 2002,2003 Matthias Kramm <kramm@quiss.org>
Copyright (c) 2002 Peter Dikant <peter.dikant@mobilexag.de>

This file is distributed under the GPL, see file COPYING for details. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "RealSpeak.h"

#define TTS_OUTPUT_BUFFER 1048576 // 1 MByte output buffer

// this header will be appended to the output created by the tts system
// to create a complete wav memory structure
static unsigned char wavHeader[] = {
	0x52, 0x49, 0x46, 0x46, 0x00, 0x00, 0x00, 0x00, // |R|I|F|F|4 byte lenght of file - 8|
	0x57, 0x41, 0x56, 0x45, 0x66, 0x6d, 0x74, 0x20, // |W|A|V|E|f|m|t| |
	0x10, 0x00, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, // |4 byte length of header|2 bytes encoding (01 = pcm)|2 bytes number of channels|
	0x40, 0x1f, 0x00, 0x00, 0x80, 0x3e, 0x00, 0x00, // |4 byte samplesrate|4 byte samplerate * bytes per sample|
	0x02, 0x00, 0x10, 0x00, 0x64, 0x61, 0x74, 0x61, // |2 byte bytes per sample|2 byte bit per sample|d|a|t|a|
	0x00, 0x00, 0x00, 0x00 };

char* RealSpeak::dictionary = 0;
char* RealSpeak::remote_server = 0;
char* RealSpeak::remote_service = 0;
int RealSpeak::remote_port = 0;
char* language = "german";
char* voice = "female1";

int RealSpeak::initialize(Logger*log)
{
    log->logf("<verbose> Using Realspeak as Text to Speech engine\n");
    return 1;
}

void RealSpeak::finalize(Logger*log)
{
    if(RealSpeak::dictionary)
	free(RealSpeak::dictionary);
}

/**
 * the default constructor will create the basic tts instances and also store a pointer
 * to the logging subsystem.
 * @param log a pointer to the logging object
 */
RealSpeak::RealSpeak(Logger *log)
    :TextToSpeech(log)
{
    log->logf("<debug> Initializing RealSpeak Engine");
    log->logf("<debug> starting with data buffer of %d bytes", TTS_OUTPUT_BUFFER);

    ttsParm.nLanguage = TTS_LANG_GERMAN;
    ttsParm.nVoice = TTS_RS_VOICE_FEMALE;

    if(!strcmp(tts_language, "german")) {
	ttsParm.nLanguage = TTS_LANG_GERMAN;
	log->logf("<notice> Setting language to LANG_GERMAN");
    } else if(!strcmp(tts_language, "benglish")) {
	ttsParm.nLanguage = TTS_LANG_BRITISH_ENGLISH;
	log->logf("<notice> Setting language to LANG_BRITISH_ENGLISH");
    } else {
	log->logf("<error> Language not known: \"%s\"", tts_language);
	log->logf("<error> Languages supported: \"german\", \"benglish\"");
	log->logf("<notice> Setting language to LANG_GERMAN");
	ttsParm.nLanguage = TTS_LANG_GERMAN;
    }

    if(!strcmp(tts_voice,"female1")) {
	ttsParm.nVoice = 0;
	log->logf("<notice> Setting voice to RS_VOICE_FEMALE");
    } else if(!strcmp(tts_voice,"female4")) {
	ttsParm.nVoice = 2;
	log->logf("<notice> Setting voice to 3000_VOICE_FEMALE");
    } else if(!strcmp(tts_voice,"female2")) {
	ttsParm.nVoice = 4;
	log->logf("<notice> Setting voice to RS_VOICE_FEMALE2");
    } else if(!strcmp(tts_voice,"female3")) {
	ttsParm.nVoice = 6;
	log->logf("<notice> Setting voice to RS_VOICE_FEMALE3");
    } else if(!strcmp(tts_voice,"male1")) {
	ttsParm.nVoice = 1;
	log->logf("<notice> Setting voice to RS_VOICE_MALE");
    } else if(!strcmp(tts_voice,"male4")) {
	ttsParm.nVoice = 3;
	log->logf("<notice> Setting voice to 3000_VOICE_MALE");
    } else if(!strcmp(tts_voice,"male2")) {
	ttsParm.nVoice = 5;
	log->logf("<notice> Setting voice to RS_VOICE_MALE2");
    } else if(!strcmp(tts_voice,"male3")) {
	ttsParm.nVoice = 7;
	log->logf("<notice> Setting voice to RS_VOICE_MALE3");
    } else {
	log->logf("<error> Voice not known: \"%s\"", tts_voice);
	log->logf("<error> Voices supported: male1, male2, male3, male4, female1, female2, female3, female4");
	log->logf("<notice> Setting voice to female1");
	ttsParm.nVoice = 0;
    }

    if(ttsParm.nLanguage != TTS_LANG_US_ENGLISH)
	/* only us-english supports different voices */
	ttsParm.nVoice = TTS_RS_VOICE_FEMALE;

    ttsParm.nOutputType = TTS_LINEAR_16BIT;
    ttsParm.nFrequency = TTS_FREQ_8KHZ;
    ttsParm.nInputDataType = TTS_DATA_TYPE_TEXT;
    ttsParm.nOutputDataType = TTS_DATA_TYPE_PCM;
    ttsParm.cbFuncs.TtsSourceCb = sourceCallback;
    ttsParm.cbFuncs.TtsDestCb = destCallback;
    ttsParm.cbFuncs.TtsEventCb = CbTtsEventNotify;
    ttsParm.cbFuncs.numCallbacks = 3;
    ttsParm.szLibLocation = ".\\Engine";

    int ret;
   
    if(RealSpeak::remote_server && RealSpeak::remote_service && RealSpeak::remote_port) {
	
	ttsParm.szLibLocation = NULL; 

	LH_SDK_SERVER* server = new LH_SDK_SERVER;
	server->server_handle = 0;
	strcpy(server->server.IP_Address, RealSpeak::remote_server);
	strcpy(server->server.service, RealSpeak::remote_service);
	server->server.port_number = RealSpeak::remote_port;
	log->logf("<notice> Connecting to remote TTS-server <%s> %s:%d", RealSpeak::remote_service,
		RealSpeak::remote_server, RealSpeak::remote_port);
	ret = TtsCreateEngine(server);
	log->logf("<notice> Connected. Handle=%d Server=%s Service=%s Port=%d", server->server_handle,
		server->server.IP_Address, server->server.service, server->server.port_number);
	if (ret != 0) {
	    log->logf("<error> Error connecting to server: %d", ret);
	}
	ret = TtsInitialize(&hInst, server, &ttsParm, (void *)this);
    } else {
	ret = TtsInitialize(&hInst, NULL, &ttsParm, (void *)this);
    }
    if (ret != 0) {
	log->logf("<error> Error initializing RealSpeak Engine: %d", ret);
    } else {
	log->logf("<debug> RealSpeak Engine initialized.");
    }
   
    if(RealSpeak::dictionary)
	loadDictionary(RealSpeak::dictionary);

    output = (unsigned char*)malloc(TTS_OUTPUT_BUFFER + 44);
    outputSize = TTS_OUTPUT_BUFFER;
    memcpy(output, wavHeader, 44);
    initMutex(&mutex);
}

TTSRETVAL RealSpeak::CbTtsEventNotify (void *pAppData, void *ppBuffer, U16 nDataSize, U16 event) {
    RealSpeak *me = (RealSpeak*)pAppData;
    //me->log->logf("<debug> Received RealSpeak Event: %d", event);
    return 0;
}

TTSRETVAL RealSpeak::sourceCallback(void *pAppData, void *data, U32 len, U32 *datasize) {
    RealSpeak *me = (RealSpeak*)pAppData;
    
    if (!me->text || !*(me->text)) {
	*datasize = 0;
	if(me->text) {
	    free(me->textstart);
	    me->text = 0;
	}
	return TTS_ENDOFDATA;
    }
    if (len >= strlen(me->text)) {
	memcpy(data, me->text, strlen(me->text));
	*datasize = strlen(me->text);
	me->text += strlen(me->text);
    } else {
	memcpy(data, me->text, len);
    	*datasize = len;
	me->text += len;
    }
    return TTS_SUCCESS;
}

void* RealSpeak::destCallback(void *pAppData, U16 nDatatype, void *data, U32 datasize, U32 *buffersize) {
    RealSpeak *me = (RealSpeak*)pAppData;
    void*ret = 0;
    //me->log->logf("<debug> TTS: DestCallback: datatype:%d data:%08x size:%d", nDatatype, data, datasize);

    while((int)(datasize+TTS_OUTPUT_BUFFER/2) > me->outputSize - me->outputPos) {
	me->outputSize += TTS_OUTPUT_BUFFER;
	void*newdata = realloc(me->output, me->outputSize + 44);
	if(!newdata) {
	    // failed, cut off
	    me->log->logf("<error> TTS: realloc(%d) failed", me->outputSize);
	    datasize = me->outputSize - me->outputPos;
	    me->outputSize -= TTS_OUTPUT_BUFFER; //revert
	    break;
	} else {
	    me->log->logf("<notice> TTS: expanded output buffer to %d", me->outputSize);
	    me->output = (unsigned char*)newdata;
	}
    }
    *buffersize = me->outputSize - me->outputPos;

    ret = (void *)&me->output[me->outputPos + 44];
    me->outputPos += (long)datasize;
    //me->log->logf("<debug> TTS: DestCallback: returning memory for %d bytes", *buffersize);
    return ret;
}

/**
 * free used ressources
 */
RealSpeak::~RealSpeak()
{
	log->logf("<debug> Deleting RealSpeak Engine");
	int ret = TtsUninitialize(hInst);
	if (ret != 0) {
	    log->logf("<error> Error closing RealSpeak Engine: %d", ret);
	}
	free(output);
	destroyMutex(&mutex);
}

/*
 * load a User Dictionary
 */
void RealSpeak::loadDictionary(const char* dictionary)
{
	int ret;
	if(dictionary) {
	    userdict = 0;
	    log->logf("<notice> Loading Realspeak dictionary \"%s\"...", dictionary);
	    ret = TtsLoadUsrDict(0, &userdict, (char*)dictionary);
	    if(ret != 0) {
		log->logf("<error> Couldn't load dictionary \"%s\", error %d/%08x", dictionary, ret, userdict);
	    } else {
		ret = TtsEnableUsrDict(hInst, userdict);
		if(ret != 0) {
		    log->logf("<error> Couldn't enable dictionary \"%s\", error %d", dictionary, ret);
		} else {
		    log->logf("<notice> Dictionary \"%s\" loaded", dictionary);
		}
	    }
	}
}

/**
 * set implementation specific parameters
 */
int RealSpeak::setParams(const char *key, const char *value)
{
	/* test which parameters are for us */
	if(!strncmp(key, "realspeak.", 10))
	{
	    const char*key2 = &key[10];
	    if(!strcmp(key2, "dictionary"))
	    {
		RealSpeak::dictionary = strdup(value);
		return 1;
	    } else if(!strcmp(key2, "server"))
	    {
		RealSpeak::remote_server = strdup(value);
		return 1;
	    }
	    if(!strcmp(key2, "service"))
	    {
		RealSpeak::remote_service = strdup(value);
		return 1;
	    }
	    if(!strcmp(key2, "port"))
	    {
		RealSpeak::remote_port = atoi(value);
		return 1;
	    }
	    if(!strcmp(key2, "language"))
	    {
		RealSpeak::language = strdup(value);
		return 1;
	    }
	    if(!strcmp(key2, "voice"))
	    {
		RealSpeak::voice = stdup(value);
		return 1;
	    }
	}
	return 0;
}

/**
 * this method will convert a given text to speech and output this over the microsoft
 * audio mapper, thus creating direct audio output on the soundcard.
 * @param text the text to be converted to speech
 */
void RealSpeak::text2Audio(const char *text)
{
	log->logf("<verbose> text2Audio is not implemented in RealSpeak");
}

/**
 * create speech directly to a wave file
 * @param text the text to be spoken
 * @param filenam name and path of the file to be created
 */
void RealSpeak::text2File(const char *text, const char *filename)
{
    lockMutex(&mutex);
    this->text = this->textstart = strdup(text);

    // start conversion
    int ret = TtsProcess(hInst);
    if (ret != 0) {
	log->logf("<error> TTS: could not process input, got error: %d", ret);
	return;
    }

    log->logf("<debug> TTS: speech complete, generated %d bytes of data.", outputPos);

    unsigned int wavelen = outputPos + 36;
    memcpy(output + 4, &wavelen, 4);
    wavelen = outputPos;
    memcpy(output + 40, &wavelen, 4);

    FILE *fp = fopen(filename, "wb");
    fwrite(output, sizeof(unsigned char), outputPos + 44, fp);
    fclose(fp);
    unlockMutex(&mutex);
}


/**
 * this method will create speech that is stored in a memory buffer. the created speech has the format pcm mono with 
 * 8kHz sampling rate and 16 bit. The memory structure can be directly used for input to the capi
 * subsystem.
 * @param text this text will be spoken
 * @return oubuffer pointer to the output buffer (the memory will be allocated)
 * @return size the size of the allocated memory
 */
void RealSpeak::text2Stream(const char *text, unsigned char **outbuffer, unsigned long *size)
{
    lockMutex(&mutex);
    this->text = this->textstart = strdup(text);

    // start conversion
    outputPos = 0;
    int ret = TtsProcess(hInst);
    if (ret != 0) {
	log->logf("<error> TTS: could not process input, got error: %d", ret);
	*size = 0;
	return;
    }

    log->logf("<debug> TTS: speech complete, generated %d bytes of data.", outputPos);

    
    *outbuffer = (unsigned char*)malloc (outputPos + 44);
    unsigned int wavelen = outputPos + 36;
    memcpy(&output[4], &wavelen, 4);
    wavelen = outputPos;
    memcpy(&output[40], &wavelen, 4);
    memcpy(*outbuffer, output, outputPos + 44);
    *size = outputPos + 44;
    
/*    FILE *fp = fopen("temp.wav", "wb");
    fwrite(*outbuffer, sizeof(unsigned char), outputPos + 44, fp);
    fclose(fp); */
    unlockMutex(&mutex);
}
