/* Svox.cpp

svox(TM) driver.

Part of the caiviar package. 

This file is distributed under the GPL, see file COPYING for details. 
*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include "Svox.h"
#include "svoxapi.h"

/** Static variables. */
char*	    Svox::dir = NULL;
char*	    Svox::voice = NULL;
int         Svox::speed = 100;
int         Svox::pitch = 100;
int         Svox::volume = 100;

/** Called only once when the capiserver is started. */
int Svox::initialize(Logger*log)
{
    // Startup SVOX
    int status;
    if ( dir == NULL ) {
	status = SVOX_Startup(".");
    } else {
	log->logf( "<verbose> Starting svox with path %s", dir );
	status = SVOX_Startup((char*)dir);
    }
    if ( status != SVOX_NoErr ) {
	return 1;
    } else {
	return 0;
    }
}

void Svox::finalize(Logger*log)
{
    log->log("<verbose> Svox::finalize()\n" );
    if(Svox::dir)
	free(Svox::dir);
    if(Svox::voice)
	free(Svox::voice);
}

/**
 * the default constructor will create the basic tts instances and also store a pointer
 * to the logging subsystem.
 * @param log a pointer to the logging object
 */
Svox::Svox(Logger *log) 
: TextToSpeech(log)
{
    log->log("<debug> Svox::Svox()");
}

/**
 * free used ressources
 */
Svox::~Svox()
{
    log->log("<verbose> Svox::~Svox()");
}

/**
 * Set implementation specific parameters
 */
int Svox::setParams(const char *key, const char *value)
{
    /* test which parameters are for us */
    if(!strncmp(key, "svox.", 5))
    {
	const char*key2 = &key[5];
	if (!strcasecmp(key, "svox.dir")) {
	    Svox::dir = strdup(value);
	    return 1;
	}
	else if (!strcasecmp(key, "svox.voice")) {
	    Svox::voice = strdup(value);
	    return 1;
	}
	else if (!strcasecmp(key, "svox.speed")) {
	    Svox::speed = atoi(value);
	    return 1;
	}
	else if (!strcasecmp(key, "svox.pitch")) {
	    Svox::pitch = atoi(value);
	    return 1;
	}
	else if (!strcasecmp(key, "svox.volume")) {
	    Svox::volume = atoi(value);
	    return 1;
	}
    }
    return 0;
}

/**
 * this method will convert a given text to speech and output this over the microsoft
 * audio mapper, thus creating direct audio output on the soundcard.
 * @param text the text to be converted to speech
 */
void Svox::text2Audio(const char *text)
{
    log->log("<verbose> text2Audio is not implemented in svox");
}

/**
 * create speech directly to a wave file
 * @param text the text to be spoken
 * @param filenam name and path of the file to be created
 */
void Svox::text2File(const char *text, const char *filename)
{
    log->log("<verbose> text2File is not implemented in svox");
}


/**
 * This method will create speech that is stored in a memory buffer. 
 *
 * Unfortunately svox does not support creating speech into a memory buffer.
 * That's why we create a file and copy the file into the memory buffer.
 * 
 * @param text this text will be spoken
 * @return oubuffer pointer to the output buffer (the memory will be allocated)
 * @return size the size of the allocated memory
 */
void Svox::text2Stream(const char *text, unsigned char **outbuffer, unsigned long *size)
{
    log->log("<verbose> text2Stream");

    const char* const filename = "sound.wav";
    SVOX_Channel chan = SVOX_NewFileChannel("",filename,SVOX_EncLin);
    if ( voice != NULL ) {
	SVOX_SetVoice(chan,(char*)voice);
	log->log((char*)voice);
    }
    SVOX_SetSpeed( chan, speed );
    SVOX_SetPitch( chan, pitch );
    SVOX_SetVolume( chan, volume );
    // do the job
    SVOX_SynthesizeString(chan,(char*)text);
    SVOX_CloseFileChannel(chan);

    FILE*fi = fopen(filename, "rb");
    if(!fi) {
	log->logf("<error> SVOX TTS failed");
	*outbuffer = 0;
	*size = 0;
	return;
    }

    // get the file size
    fseek(fi, 0, SEEK_END);
    int l = ftell(fi);
    fseek(fi, 0, SEEK_SET);
    if(!l) {
	log->logf("<error> SVOX TTS generated no output");
	*outbuffer = 0;
	*size = 0;
	return;
    }

    // allocate memory buffer and read in the generated sound
    *outbuffer = (unsigned char*)malloc(l);
    *size = l;
    fread(*outbuffer, l, 1, fi);
    fclose(fi);
    unlink(filename);
}
